"""SCRU-160: Sortable, Clock and Random number-based Unique identifier"""

import datetime
import secrets
import threading
import typing

__all__ = ["scru160", "scru160f"]


class Generator:
    """Represents SCRU-160 ID generator."""

    def scru160(self) -> str:
        """Generates a new SCRU-160 ID encoded in the base32hex format.

        Returns:
            32-character base32hexupper string (/^[0-9A-V]{32}$/).
        """
        return _base32hex160(self.generate())

    def scru160f(self) -> str:
        """Generates a new SCRU-160 ID encoded in the hexadecimal format.

        Returns:
            40-character hexadecimal string (/^[0-9a-f]{40}$/).
        """
        return self.generate().hex()

    def generate(self) -> bytes:
        """Generates a byte sequence that represents a new SCRU-160 ID.

        Returns:
            20-byte sequence.
        """
        ts, cnt = self._get_ts_and_cnt()
        return (
            ts.to_bytes(6, "big")
            + cnt.to_bytes(2, "big")
            + secrets.randbits(96).to_bytes(12, "big")
        )

    def __init__(self) -> None:
        self._ts: int = 0
        self._cnt: int = 0
        self._lock: threading.Lock = threading.Lock()

    def _get_ts_and_cnt(self) -> typing.Tuple[int, int]:
        """Updates the internal state and returns the latest values.

        Returns:
            The latest timestamp and counter.
        """

        def now() -> int:
            return int(datetime.datetime.now().timestamp() * 1000)

        with self._lock:
            new_ts = now()
            if new_ts <= self._ts:
                self._cnt += 1
                if self._cnt < 0x1_0000:
                    return (self._ts, self._cnt)

                # wait a moment until clock goes forward
                i = 0
                while new_ts <= self._ts:
                    new_ts = now()
                    i += 1
                    if i > 1_000_000:
                        import warnings

                        warnings.warn(
                            "scru160: reinitialized internal state as clock did not go forward; monotonicity may be broken",
                            RuntimeWarning,
                        )
                        break

            self._ts = new_ts
            self._cnt = secrets.randbits(15)
            return (self._ts, self._cnt)


_default_generator = Generator()


def scru160() -> str:
    """Generates a new SCRU-160 ID encoded in the base32hex format.

    Returns:
        32-character base32hexupper string (/^[0-9A-V]{32}$/).
    """
    return _default_generator.scru160()


def scru160f() -> str:
    """Generates a new SCRU-160 ID encoded in the hexadecimal format.

    Returns:
        40-character hexadecimal string (/^[0-9a-f]{40}$/).
    """
    return _default_generator.scru160f()


def _base32hex160(bs: bytes) -> str:
    """Encodes 20-byte sequence into base32hex."""
    cs = "0123456789ABCDEFGHIJKLMNOPQRSTUV"
    buffer = []
    for i in [0, 5, 10, 15]:
        buffer += [
            cs[31 & (bs[i + 0] >> 3)],
            cs[31 & (bs[i + 0] << 2) | (bs[i + 1] >> 6)],
            cs[31 & (bs[i + 1] >> 1)],
            cs[31 & (bs[i + 1] << 4) | (bs[i + 2] >> 4)],
            cs[31 & (bs[i + 2] << 1) | (bs[i + 3] >> 7)],
            cs[31 & (bs[i + 3] >> 2)],
            cs[31 & (bs[i + 3] << 3) | (bs[i + 4] >> 5)],
            cs[31 & (bs[i + 4] >> 0)],
        ]

    return "".join(buffer)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scru160']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['scru160 = scru160.cli:generate']}

setup_kwargs = {
    'name': 'scru160',
    'version': '0.2.1',
    'description': 'SCRU-160: Sortable, Clock and Random number-based Unique identifier',
    'long_description': '# SCRU-160: Sortable, Clock and Random number-based Unique identifier\n\nSCRU-160 ID is yet another attempt to supersede [UUID] in the use cases that\nneed decentralized, globally unique time-ordered identifiers. SCRU-160 is\ninspired by [ULID] and [KSUID] and has the following features:\n\n- 160-bit length\n- Sortable by generation time (in binary and in text)\n- Two case-insensitive encodings: 32-character base32hex and 40-character hex\n- More than 32,768 unique, time-ordered but unpredictable IDs per millisecond\n- Nearly 111-bit randomness for collision resistance\n\n```python\nfrom scru160 import scru160, scru160f\n\nprint(scru160())  # e.g. "05TVFQQ8UGDNKHDJ79AEGPHU7QP7996H"\nprint(scru160())  # e.g. "05TVFQQ8UGDNNVCCNUH0Q8JDD3IPHB8R"\n\nprint(scru160f())  # e.g. "017bf7eb48f41b7d6bd295bc5adc43436bc969df"\nprint(scru160f())  # e.g. "017bf7eb48f41b7e1bc98aec348dfa1539b41288"\n```\n\n[uuid]: https://en.wikipedia.org/wiki/Universally_unique_identifier\n[ulid]: https://github.com/ulid/spec\n[ksuid]: https://github.com/segmentio/ksuid\n\n## License\n\nCopyright 2021 LiosK\n\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at\n\nhttp://www.apache.org/licenses/LICENSE-2.0\n\nUnless required by applicable law or agreed to in writing, software distributed\nunder the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR\nCONDITIONS OF ANY KIND, either express or implied. See the License for the\nspecific language governing permissions and limitations under the License.\n',
    'author': 'LiosK',
    'author_email': 'contact@mail.liosk.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/scru160/python',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.7.12,<4.0.0',
}


setup(**setup_kwargs)



"""Command-line interface to generate SCRU-160 identifiers"""
import argparse

from . import scru160, scru160f


def generate() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f", action="store_true", help="print identifiers in hex encoding"
    )
    parser.add_argument(
        "-n",
        default=1,
        type=int,
        help="generate given number of identifiers",
        metavar="count",
    )

    args = parser.parse_args()
    fn = scru160f if args.f else scru160
    for _ in range(args.n):
        print(fn())
